//"use strict";

var g_version = "v0.7.5";

var EMPTY = 0;
var WHITE = 32;
var BLACK = 64;
var TYPE = 31;
var ROYAL = 256;

// id for SP_* functions.
var SOLVER = 0;
var VIEWER = 1;

var pieceLetters = {
	englishN: ["P", "N", "B", "R", "Q", "K", "H", "U",
	           "F",  "W",  "M",  "A",  "C",  "V",
	           "J",  "E",  "D",  "S",  "X"],

	englishS: ["P", "S", "B", "R", "Q", "K", "N", "BP",
	           "FE", "WE", "EK", "PR", "EM", "AM",
	           "WA", "ET", "DR", "SH", "X"],

	french:   ["P", "C", "F", "T", "D", "R", "N", "PB",
	           "FE", "VI", "RE", "PR", "IM", "AM",
	           "V",  "EL", "DR", "BA", "X"],

	german:   ["B", "S", "L", "T", "D", "K", "N", "BB",
	           "FE", "WE", "EK", "PR", "KN", "AM",
	           "WA", "ET", "DR", "SC", "X"]

// https://pdb.dieschwalbe.de/pieces.jsp
//	germanPDB: ["B", "S", "L", "T", "D", "K", "N", "BE",
//	            "FE", "WE", "EK", "PR", "KA", "AM",
//	            "WA", "ET", "DC", "SH", "X"]
};

var pieceFigurines = {
	// Use unicode characters, and not escaped html, so that these
	// strings work in a textarea.
	// Someday we can use the new unicode block for fairy pieces,
	// not very supported yet.
	// https://en.wikipedia.org/wiki/Chess_Symbols_(Unicode_block)
	w: ["\u2659", "\u2658", "\u2657", "\u2656", "\u2655", "\u2654",
	    "wN", "wBP",
	    "wFE", "wWE", "wEK", "wPR", "wEM", "wAM",
//	    "\uD83E\uDE22", "\uD83E\uDE23",	// N, BP
//	    "\uD83E\uDE36", "\uD83E\uDE35", "\uD83E\uDE34",	// FE, WE, EK
//	    "\uD83E\uDE50", "\uD83E\uDE4F", "\uD83E\uDE4E",	// PR, EM, AM
	    "wET", "wWA", "wDR", "wSH", "\u25CB"],
	b: ["\u265F", "\u265E", "\u265D", "\u265C", "\u265B", "\u265A",
	    "bN", "bBP",
	    "bFE", "bWE", "bEK", "bPR", "bEM", "bAM",
//	    "\uD83E\uDE28", "\uD83E\uDE29",
//	    "\uD83E\uDE3C", "\uD83E\uDE3B", "\uD83E\uDE3A",
//	    "\uD83E\uDE53", "\uD83E\uDE52", "\uD83E\uDE51",
	    "bET", "bWA", "bDR", "bSH", "\u25CF"]
};

function languageObject(lang) {
	if (!lang)
		lang = $("select#output-language").val();

	if (lang == "englishN") {
		return {
			lang: lang,
			letters: pieceLetters.englishN,
			colors: ["w", "b"],
			keywords: ["pieces", "white", "black", "royal"]
		};
	}
	else if (lang == "englishS" || lang == "py2web") {
		return {
			lang: lang,
			letters: pieceLetters.englishS,
			colors: ["w", "b"],
			keywords: ["pieces", "white", "black", "royal"]
		};
	}
	else if (lang == "french" || lang == "french2") {
		return {
			lang: lang,
			letters: pieceLetters.french,
			colors: ["b", "n"],
			keywords: ["pieces", "blanc", "noir", "royale"]
		};
	}
	else if (lang == "german") {
		return {
			lang: lang,
			letters: pieceLetters.german,
			colors: ["w", "s"],
			keywords: ["steine", "weiss", "schwarz", "koeniglich"]
		};
	}
	else if (lang == "figurine") {
		var UNDEFINED_PIECE = pieceFigurines.w.length - 1;
		return {
			lang: lang,
			letters: pieceFigurines.w,	// To render row960 in figurine.
			colors: [pieceFigurines.w[UNDEFINED_PIECE],
			         pieceFigurines.b[UNDEFINED_PIECE]],
			keywords: ["pieces", "white", "black", "royal"]
		};
	}
	else {
		openDialog("Error", "Unknown output language '" + lang + "'.");
		// Return something sensible to limit the error.
		return {
			lang: lang,
			letters: pieceLetters.englishS,
			colors: ["w", "b"],
			keywords: ["pieces", "white", "black", "royal"]
		};
	}
}

var DEBUG = false;	// To turn on some console.log calls.

// https://stackoverflow.com/questions/15337594/prevent-scroll-of-the-page-if-pop-up-is-open
$(document).on("popupafteropen", "div#popupDialog", function(event, ui) {
	$("div#popupDialog a").focus();
	$("body").css("overflow", "hidden").on("touchmove", function(e) {
		e.preventDefault();
	});
}).on("popupafterclose", "div#popupDialog", function(event, ui) {
	$("body").css("overflow", "auto").off("touchmove");
	$("div#popupDialog h1").html("");	// Clear for better debugging.
	$("div#popupDialog p").html("");	// Clear because we're appending.
});

// https://stackoverflow.com/questions/6234773/can-i-escape-html-special-chars-in-javascript
function escapeHtml(unsafe) {
	return unsafe
		.replace(/&/g, "&amp;")
		.replace(/</g, "&lt;")
		.replace(/>/g, "&gt;")
		.replace(/"/g, "&quot;")
		.replace(/'/g, "&#039;");
}

// Helper functions.
function openDialog(title, text) {
	$("div#popupDialog h1").html(title);

	var $p = $("div#popupDialog p");
	if ($p.html() == "")
		$p.html(escapeHtml(text));
	else	// In case of subsequent calls.
		$p.append("<br>" + escapeHtml(text));

	$("div#popupDialog").popup("open");
}

// ---------------------------------------------------------------------------

// Functions from engine. These are undefined for now, and will get set
// asynchronously. They will be available when g_engineReady = true.

var SetProblemFromJSON;
var GetLastPly;
var SP_PlayMove;
var SP_ReverseMove;
var SP_GetLastMoveNotation;
var Solver_NonsenseCombination;
var Solver_CanCaptureTheKing;
var Solver_PrepareDemolitionPass;
var Solver_IsPassIncomplete;
var Solver_RunStep;
var Solver_SearchFraction;
var Solver_GetRow960;
var TT_LoadFraction;
var Viewer_SetRow960;
var Viewer_ComputeLegalMoves;
var Viewer_GetMoveLength;
var Viewer_HeuristicMoveType;
var Viewer_GetSquare;

// ---------------------------------------------------------------------------

// General.

var g_params;
var g_useWasm;

// Problem editor.

var g_commands = [];
var g_language = "englishS";
var g_load_request = null;

// Solver.

var g_hashMem;
var g_engineLaunched = false;
var g_engineReady = false;

var g_firstPly = 0;
var g_lastPly = 0;
var g_initialMoveNumber = 1;

var g_abortMessage = "";	// Set to "get" to request abort messages.
var g_paused = false;
var g_lastCurrTime = null;
var g_workChunk = 0;

var g_computeTime = 0;
var g_pass = 0;
var g_last_extracted = null;
var g_exploring = {row960: "", moves: []};
var g_longestAttempt = {row960: "", moves: []};
var g_solutions = [];
var g_solutionsNum = 0;	// Different from g_solutions.length if "count only".
var g_solutionsLastLen = 0;	// Number of solutions displayed in the UI.

var g_nodesBatch = 0;
var g_nodesCount = 0;
var g_nodesComputeTime = 0;
var g_nodesSpeed = "?";
var g_heuristicMode = false;
var g_soln_type = "";
var g_editedSpan = $();	// Empty jQuery object.
var g_verticalCylinder = false;

// Game editor.

var g_editorReady = false;
var g_editedGame = {row960: "", moves: [], reject: "", str960: ""};
var g_editedDepth = 0;	// Current depth in g_editedGame.
var g_legalMoves = "";	// String containing legal moves in current diagram.

// 0: waiting for origin mousedown
// 1: waiting for origin mouseup
// 2: waiting for destination mouseup.
var g_moveState = 0;

var g_originSquare = "";
var g_originPiece = null;	// Will catch two divs if royal.
var g_mouseDownX = 0;// To distiguish between actual move and mouse vibration.
var g_mouseDownY = 0;//
var g_mouseMoveSeen = false;

var g_createdDepth = null;
var g_createStip1 = "dia";
var g_createStip2 = "a=>b";

// ---------------------------------------------------------------------------

// Parse hash string for parameters.
g_params = {};
if (window.location.hash.substr(0, 8) == "#solver?") {
	var str = window.location.hash.substr(8);
	var parts = str.split("&");
	for (var i = 0; i < parts.length; i++) {
		var j = parts[i].indexOf("=");
		var key;
		var value;
		if (j == -1) {
			key = parts[i];
			value = "";
		}
		else {
			key = parts[i].substr(0, j);
			value = parts[i].substr(j + 1);
		}
		g_params[key] = unescape(value);
	}
}

// Disable wasm for early versions of iOS.
// Here are some data points:
// 12.4 (2019-07-22): broken
// 13.0 (2019-09-19): works
// 13.1.3 (2019-10-15): works.
// It's probably this:
// Safari Technology Preview Release 86 (Jun 26, 2019)
// - Fixed float64 "select" on ARM64 devices
// Covers WebKit revisions 246093-246691, tags/Safari-608.1.30.1
// iOS version 12.4.1 WebKit version 607.1.40
// iOS version 13.0   WebKit version 608.2.11

// Also disable wasm for
// location.protocol == "file:" for every browser excepted Firefox <68.
// https://developer.mozilla.org/en-US/docs/Web/HTTP/CORS/Errors/CORSRequestNotHttp

var disableWasm = (bowser.osname == "iOS" &&
                   (typeof bowser.osversion == "undefined" ||
                    bowser.compareVersions([bowser.osversion, "13.0"]) < 0)) ||
                  (location.protocol == "file:" &&
                   (bowser.name != "Firefox" ||
                    typeof bowser.version == "undefined" ||
                    bowser.compareVersions([bowser.version, "68"]) >= 0));

g_useWasm = typeof WebAssembly == "object" && !("asmjs" in g_params) &&
            (!disableWasm || ("wasm" in g_params));

// ---------------------------------------------------------------------------

$(document).one("pagecreate", "#about", function() {
	var $page = $(this);	// "this" is the page for pagecreate.

	var browser = "(unknown browser)";
	if (typeof bowser.name != "undefined") {
		browser = bowser.name;
		if (typeof bowser.version != "undefined")
			browser += " " + bowser.version;
		else
			browser += " (unknown version)";
	}

	var os = "(unknown OS)";
	if (typeof bowser.osname != "undefined") {
		os = bowser.osname;
		if (typeof bowser.osversion != "undefined")
			os += " " + bowser.osversion;
		else
			os += " (unknown version)";
	}

	var device = "desktop/laptop";
	if (bowser.mobile)
		device = "mobile";
	else if (bowser.tablet)
		device = "tablet";

	var wasm = typeof WebAssembly == "object";
	var appcache = "applicationCache" in window;
	var service_workers = "serviceWorker" in navigator;
//	var hist = "replaceState" in window.history;

	$page.find("span#browser").text(browser);
	$page.find("span#os").text(os);
	$page.find("span#device").text(device);
	$page.find("span#webassembly").text(
		(wasm ? "yes" : "no") +
		(wasm && !g_useWasm ? " (off)" : ""));
	$page.find("span#appcache").text(appcache ? "yes" : "no");
	$page.find("span#service-workers").text(service_workers ? "yes" : "no");
//	$page.find("span#history").text(hist ? "yes" : "no");
});

$(document).one("pagebeforecreate", "#doc", function() {
	$("div#doc div[data-role='header']").prepend(
		'<a href="javascript:history.back()" data-icon="carat-l">Back</a>');

	$("pre.load").prepend("<button class='ui-btn ui-btn-inline " + 
	                      "ui-corner-all load'>Load in solver</button>");

	$("button.load").click(function(){
    	var $pre = $(this).parent();
		var text = $pre.contents().get(1).nodeValue;
		// Trim any trailing newline.
		if (text.charAt(text.length - 1) == "\n")
			text = text.substr(0, text.length - 1);
		g_load_request = text;
//		$.mobile.back();	// A second way to do things, more debugging.
		window.history.back();
	});
});

// ---------------------------------------------------------------------------

// One em = 16px if !wantMini().
// One em = 12.5px if wantMini().

function wantMini() {
	return ($(window).width() < 560);
}

// Call this initially and on resize.
function adjustSizes($scope) {
	// Can't set elements. Have to set an enclosing tag.
	$scope.find("div.ui-content").toggleClass("ui-mini", wantMini());
	$scope.find("div[data-role='controlgroup']").toggleClass(
		"ui-mini", wantMini());
	$scope.find("ul[data-role='listview']").toggleClass("ui-mini", wantMini());
	$scope.find("ul.ui-listview").toggleClass("ui-mini", wantMini());
	$scope.find("fieldset").toggleClass("ui-mini", wantMini());

	// Only update board if it contains divs, otherwise the call will
	// crash if the engine is not initialized. The board itself is resized in
	// css.
	if ($scope.find("td#game-board > div").length > 0)
		updateBoard();
}

$(document).on("pagebeforecreate", function(event, ui) {
	// Don't use $.mobile.activePage which refers to the page we're
	// transitioning from.
	var $page = $(event.target);

	// Must be done on pagebeforecreate otherwise it doesn't work.
	// Sizes end up being fixed after that and don't adjust when resizing.
	adjustSizes($page);
});

$(window).resize(function() {
	adjustSizes($("body"));
});
